from typing import Union, List, Set, Dict, Optional
from typeguard import typechecked
import abc

from rstt import Player, Duel, BetterWin
from rstt.stypes import Solver, Achievement
from rstt.ranking.ranking import Ranking
import rstt.utils.utils as uu

from collections import defaultdict



class Competition(metaclass=abc.ABCMeta):
    '''
        NOTE: In the future the competition class could evolve.
            - inherit from a Scheduler class
            - composition over inheritance:
                * PlayerManager
                * GameManager
                * EventManager
    '''
    @typechecked
    def __init__(self, name: str,
                 seeding: Ranking,
                 solver: Solver=BetterWin(),
                 cashprize: Optional[Dict[int, float]]=None):
        # a name
        self.name = name
        
        # 'settings'
        self.participants = []
        self.seeding = seeding
        self.solver = solver
        self.cashprize = defaultdict(lambda: 0)
        if cashprize:
            self.cashprize.update(cashprize)
        
        # result related variable
        self.played_matches = []
        self.standing = {}
        
        # control variable
        self.__started = False
        self.__finished = False
        self.__closed = False

    # --- getter --- #
    def started(self):
        return self.__started
    
    def live(self):
        return self.__started and not self.__finished
    
    def over(self):
        return self.__closed
    
    @typechecked
    def games(self, by_rounds=False):
        return self.played_matches if by_rounds else uu.flatten(self.played_matches)
    
    @typechecked
    def top(self, place: Optional[int]=None) -> Union[Dict[int, List[Player]], List[Player]]:
        if place:
            return [key for key, value in self.standing.items() if value == place]
        else:
            return {v: [key for key, value in self.standing.items() if value == place] for v in self.standing.values()}

    # --- general mechanism --- #
    @typechecked
    def registration(self, players: Union[Player, List[Player], Set[Player]]):
        if not self.__started:
            playerset = set(self.participants)
            playerset.update(players)
            self.participants = list(playerset)

    def run(self):
        if self.__started:
            msg = f"Can not run an event that has already started. Did you mean to use play() or perhaps did you wrongly call start()?"
            raise RuntimeError(msg)
        else:
            self.start()
            self.play()
            self.trophies()

    def start(self):
        if not self.__started:
            self.seeding = self.seeding.fit(self.participants)
            self._initialise()
            self.__started = True

    def play(self):
        if not self.__started:
            msg = f"Can not play an event that has not yet started. Did you mean to use .run() or perhaps did you forgot to call .start() first?"
            raise RuntimeError(msg)
        while not self.__finished:
            current_round = self.generate_games()
            results = self.play_games(current_round)
            self.__finished = self.edit(results)

    def play_games(self, games: List[Duel]):
        played = []
        for game in games:
            self.solver.solve(game)
            played.append(game)
        return played

    def edit(self, games: List[Duel]):
            self.played_matches.append(games)
            self._update()
            return self._end_of_stage()

    def trophies(self):
        self.standing = self._standing()
        for player in self.participants:
            try: 
                result = Achievement(self.name, self.standing[player], self.cashprize[self.standing[player]])
                player.collect(result)
            except AttributeError:
                continue
        self.__closed= True

    # --- subclass specificity --- #
    def _initialise(self) -> None:
        '''Function called once, after seedings computation but before any game is played.'''
    
    def _update(self) -> None:
        '''This function is called at the end of every 'rounds', after the game have been stored, but before checking the competition end condition.'''
    
    @abc.abstractmethod
    def _end_of_stage(self) -> bool:
        '''Test if the competition should stop.'''
    
    @abc.abstractmethod
    def _standing(self) -> Dict[Player, int]:
        '''Function called once after every game is played. Builds the final standing of the event'''

    @abc.abstractmethod
    def generate_games(self) -> List[Duel]:
        '''Function called every 'round' to generate games. Should return games WITHOUT scores assigned'''
            
    


from dataclasses import dataclass
from typing import List,Union,  Any, Iterable, Protocol, runtime_checkable


# TODO: define 'Event', 'Tournament', 'MatchMaking'





class SPlayer(Protocol):
    def name(self) -> str:
        ...
    
    def level(self) -> float:
        ...


Score = list[float] # type Score = list[float] -> SYNTAX VALID SINCE 3.12, package build for 3.10


@runtime_checkable
class SMatch(Protocol):
    def players(self) -> List[SPlayer]:
        ...
    
    def scores(self) -> Score:
        ...
    
    # FIXME: def __set_result(self, score: Score) -> None:


@dataclass
class Achievement:
    event_name: str
    place: int
    prize: float
    # ??? categorie (world event, domestic league, playoffs etc.)
    # ??? event_type (SEB, Snake, MM etc)
    # ??? teamates

    
@runtime_checkable
class Solver(Protocol):
    def solve(self, match: SMatch, *agrs, **kwargs) -> None:
        ...


# -------------------------- #
# --- Typing for Ranking --- #
# -------------------------- #
@runtime_checkable
class Inference(Protocol):
    # NOTE: name source: https://en.wikipedia.org/wiki/Statistical_inference
    def rate(self, *args, **kwargs) -> Any:
        ...


@runtime_checkable        
class RatingSystem(Protocol):
    def get(self, key: SPlayer) -> Any:
        ...
        
    def set(self, key: SPlayer, rating: Any) -> None:
        ...
        
    def ordinal(self, rating: Any) -> float:
        ...


@runtime_checkable
class Observer(Protocol):
    def handle_observations(self, infer: Inference, datamodel: RatingSystem, *args, **kwargs) -> None:
        ...


# ---------------------------- #
# --- Typing for Shceduler --- #
# ---------------------------- #
@runtime_checkable
class Shuffler(Protocol):
    def rearange(status: List[int]) -> List[int]:
        """reorder elements in a 'meaningfull' fashion

        Parameters
        ----------
        status : List[int]
            A list of integers [e0, e1, e2, ...]

        Returns
        -------
        List[int]
            A list of the same integers, in a different order.
            This new ordering should be deterministic and not random. 
        """
        ...


@runtime_checkable
class Seeder(Protocol):
    def seed(players: List[SPlayer], initial_seeds: Iterable, results: Any, **kwargs) -> List[SPlayer]:
        """reorder players in a 'meaningfull' fashion

        Parameters
        ----------
        players : List[Player]
            A list of Player to order
        initial_seeds : Iterable
            An original ordering of the players
        results : Any
            Any data structure containing games (type ´Game´) related to the involved players

        Returns
        -------
        List[Player]
            A list of the same players, in a different order
        """
        ...


@runtime_checkable
class Generator(Protocol):
    def generate(status: Union[List[int], List[SPlayer]]) -> Union[List[int], List[SPlayer]]:
        """Generate different ordering version of a given List
        
        
        the different version produced are refered as 'options'
        For short input, this function could just produce all possible orderings.
        For long inout, this function could implement greedy strategy.
        Thus the input is a list, and the initial ordering is meaningfull for the strategy
        
        Parameters
        ----------
        status : Union[List[int], List[Player]]
            A list of element to compute options

        Returns
        -------
        Union[List[int], List[Player]]
            the options, A list of different 'option' which are reordering of the given 'status'
            Idealy the first option is a direct, and 'obvious' transformation of status
        """
        ...
    

@runtime_checkable
class Evaluator(Protocol):
    def eval(options: List[List[SPlayer]], initial_seeds: Iterable, results: Any, **kwargs) -> List[List[SPlayer]]:
        """reorder options based on criteria

        The options can be evaluated based on an history of Game(s), or Player(s) appreciation.
        Usuallythe eval function calls somehow somwhere a 'sort'

        Parameters
        ----------
        options : List[List[Player]]
            _description_
        initial_seeds : Iterable
            _description_
        results : Any
            _description_

        Returns
        -------
        List[List[Player]]
            _description_
        """
        ...




import random



# -------------------- #
# --- Player cfg ----- #
# -------------------- #

PLAYER_GAUSSIAN_MU = 1500
PLAYER_GAUSSIAN_SIGMA = 500

PLAYER_DIST = random.gauss
PLAYER_DIST_ARGS = {'mu': PLAYER_GAUSSIAN_MU,
                    'sigma': PLAYER_GAUSSIAN_SIGMA}

# -------------------- #
# --- Solver cfg ----- #
# -------------------- #

LOGSOLVER_BASE = 10
LOGSOLVER_LC = 400


from typing import List
from typeguard import typechecked

import abc

from rstt.player import Player

import random

class PlayerTVS(Player, metaclass=abc.ABCMeta):
    def __init__(self, *args, **kwars):
        super().__init__(*args, **kwars)
        self.__current_level = self._Player__level
        self.__level_history = []
        self.__params = kwars # level evolution parameters
        
    # --- getter --- #
    def level_history(self) -> List[float]:
        return self.__level_history
    
    def original_level(self) -> float:
        return self._Player__level
    
    def params(self):
        return self.__params
    
    # --- override --- #
    def level(self):
        return self.__current_level
    
    # --- internal mechanism --- #
    @abc.abstractmethod
    def update_level(self, *args, **kwars) -> None:
        '''change the self.__current_level value'''


class GaussianPlayer(PlayerTVS):
    @typechecked
    def __init__(self, name: str, mu: float, sigma: float):
        # pass mu as level to Player, and mu/sigam as params to PlayerTVS
        super().__init__(name, mu, mu=mu, sigma=sigma)
        self.__mu = mu
        self.__sigma = sigma
        
    def update_level(self, *args, **kwars) -> None:
        self._PlayerTVS__current_level = random.gauss(self.__mu, self.__sigma)


from typing import Union, List, Dict, Tuple, Optional, Any, Iterable
from typing import Protocol, runtime_checkable
from typeguard import typechecked

from rstt import Player, Duel, BetterWin
from rstt.ranking.ranking import Ranking
from rstt.stypes import Solver, Shuffler, Seeder, Generator, Evaluator
from rstt.scheduler.tournament.competition import Competition

from rstt.utils import utils as uu, matching as um, competition as uc


class DummyParam(Shuffler, Seeder, Evaluator):
    def rearange(status: List[int]) -> List[int]:
        return status

    def seed(players: List[Player], *args,  **kwargs) -> List[Player]:
        return players

    def generate(status: Union[List[int], List[Player]]) -> Union[List[int], List[Player]]:
        return um.ruban(status)
    
    def eval(options: List[List[Player]], *args,  **kwargs) -> List[List[Player]]:
        return options

class SwissBracket(Competition):
    @typechecked
    def __init__(self, name: str, seeding: Ranking, solver: Solver = BetterWin(),
                 matchings: Optional[Dict[Tuple[int, int], Shuffler]]=None,
                 seeders: Optional[Dict[Tuple[int, int], Seeder]]=None,
                 generators: Optional[Dict[Tuple[int, int], Generator]]=None,
                 evaluators: Optional[Dict[Tuple[int, int], Evaluator]]=None,
                 def_matching=DummyParam,
                 def_seeder=DummyParam,
                 def_generator=DummyParam,
                 def_evaluator=DummyParam,
                 cashprize: Dict[int, float] = {}):
        super().__init__(name, seeding, solver, cashprize)

        # general settings of the system
        self.current_round = 0
        self.max_round = None
        self.max_wins = None
        self.max_loses = None
    
        # round parameters - map score with data: (int, int) -> data
        self.rounds = {} # List[Player]
        self.matchings = matchings if matchings else {} # Callable
        self.seeders = seeders if seeders else {} # Callable
        self.generators = generators if generators else {} # Callable
        self.evaluators = evaluators if evaluators else {} # Callable
        
        self.def_matching = def_matching
        self.def_seeder = def_seeder
        self.def_generator = def_generator
        self.def_evaluator = def_evaluator
                
    # --- Override ---#
    def _initialise(self):
        # !!! Currently hardcoded setting for 16 participants
        msg = f"Currently SwissBracket is implemented only for 16 players, received {len(self.participants)}"
        assert len(self.participants) == 16, msg
        
        self.__init_settings()        
        self.__init_params()
        
    def _update(self) -> None:
        # ??? clear self.rounds or keep history 
        # move winner and loser
        for game in self.played_matches[-1]:
            p1, p2 = game.player1(), game.player2()
            self.rounds[self.score(p1)].append(p1)
            self.rounds[self.score(p2)].append(p2)
            
        # move to next round
        self.current_round += 1

    def _end_of_stage(self) -> bool:
        return self.current_round == self.max_round
    
    def _standing(self) -> Dict[Player, int]:
        standing = {}
        top8 = [(self.max_wins, i) for i in range(self.max_loses)]
        bottom8 = [(i, self.max_loses) for i in range(self.max_wins-1, -1, -1)]
        final_scores = top8 + bottom8
        top = 0
        for score in final_scores:
            players = self.rounds[score]
            top += len(players)
            for p in players: 
                standing[p] = top
        return standing

    def generate_games(self):
        games = [uc.find_valid_draw(draws=self.draws(group=self.rounds[score]),
                                    games=self.games()) 
                                    for score in self.round_scores()]
        return uu.flatten(games)

    '''def generate_games(self):
        try:
            games = [uc.find_valid_draw(draws=self.draws(group=self.rounds[score]),
                                        games=self.games()) 
                                        for score in self.round_scores()]
        except RuntimeWarning:
            warnings.warn('cacth and throw', RuntimeWarning)
        return uu.flatten(games)'''
    
    # --- round mechanism --- #
    @typechecked
    def score(self, player: Player) -> Tuple[int, int]:
        wins = len([game for game in self.games() if game.winner() == player])
        loses = len([game for game in self.games() if game.loser() == player])
        return (wins, loses)

    def round_scores(self) -> List[Tuple[int, int]]:
        # Round1 [(0, 0)]; Round2 [(1,0), (0,1)]; Round3 [(2,0), (1,1), (0,2)] ...
        return [(self.current_round-i, i) for i in range(self.current_round+1)
                if i < self.max_loses and self.current_round-i < self.max_wins]

    @typechecked
    def draws(self, group: List[Player]) -> List[List[Duel]]:
        scores = [self.score(player) for player in group]
        assert len(set(scores)) == 1 # !!! debugging - SwissBracket will change to support different amount of players
        score = scores[0]        
        
        # sellect correct round parameters
        matching = self.matchings[score].rearange
        seeder = self.seeders[score].seed
        generator = self.generators[score].generate
        evaluator = self.evaluators[score].eval

        # how do seeds face each other by default
        paired_list = matching(list(range(len(group))))
        
        # who has which seeds - replace 'int' by 'Player'
        paired_player = seeder(players=[group[i] for i in paired_list],
                               initial_seeds=self.seeding,
                               results=self.results())
        
        # what are all the options we want to check
        options = generator(paired_player)
       
        # in which order should we consider them
        options = evaluator(options,
                            initial_seeds=self.seeding,
                            results=self.results())
        
        return[uc.playersToDuel(option) for option in options]

    def results(self):
        # fit data to expected format for tiebreaker function
        res = {player: {'opponent': [], 'score': [],} for player in self.seeding}
       
        for game in self.games():
            # update opponent list
            res[game.player1()]['opponent'].append(game.player2())
            res[game.player2()]['opponent'].append(game.player1())
            
            # update score list !!! Lacks generalisation
            res[game.player1()]['score'].append('win' if game.winner() == game.player1() else 'lose')
            res[game.player2()]['score'].append('win' if game.winner() == game.player2() else 'lose')
        return res
    
    # --- under the hood mechanism --- #    
    def __possible_scores(self):
        return [(i,j) for j in range(self.max_loses+1) for i in range(self.max_wins+1) if i+j <= self.max_round]

    def __init_settings(self):
        self.max_round = 5
        self.max_wins = 3
        self.max_loses = 3
        
    def __init_params(self):
        scores = self.__possible_scores()
        
        self.rounds[(0, 0)] = [player for player in self.seeding]
        self.rounds = {score: self.rounds[score] if score in self.rounds.keys()
                       else [] for score in scores}
        
        self.matchings = {score: self.matchings[score] if score in self.matchings.keys()
                                  else self.def_matching for score in scores}
        
        self.seeders = {score: self.seeders[score] if score in self.seeders.keys()
                                  else self.def_seeder for score in scores}
        
        self.generators = {score: self.generators[score] if score in self.generators.keys()
                                  else self.def_generator for score in scores}
        
        self.evaluators = {score: self.evaluators[score] if score in self.evaluators.keys()
                                  else self.def_evaluator for score in scores}